from typing import Dict, List

from PySide2.QtWidgets import QWidget, QVBoxLayout, QWidgetItem, QLayout
from traitlets import HasTraits, List as TList, Instance, directional_link

from regexport.model import AppState
from regexport.views.labelled_slider import LabelledSliderView, LabelledSliderModel
from regexport.views.utils import HasWidget


class ChannelFilterModel(HasTraits):
    sliders = TList(Instance(LabelledSliderModel))

    def __getitem__(self, label) -> LabelledSliderModel:
        for slider in self.sliders:
            if slider.label == label:
                return slider
        raise KeyError(f"Slider with label {label} not found.")

    def register(self, model: AppState):
        self.model = model
        directional_link(
            (model, 'max_numspots_filters'),
            (self, 'sliders'),
            lambda max_filters: self.create_new_sliders(max_filters),
        )

    def set_max(self, channel: str, value: int):
        self[channel].value = value

    def create_new_sliders(self, max_numspots_filters: Dict[str, int]) -> List[LabelledSliderModel]:
        if set(slider.label for slider in self.sliders) != set(max_numspots_filters):
            sliders = []
            for chan, v in max_numspots_filters.items():
                slider = LabelledSliderModel(label=chan, max=v, value=v)
                slider.observe(self._update_model)
                sliders.append(slider)
            return sliders
        return self.sliders

    def _update_model(self, change):
        self.model.max_numspots_filters = {slider.label: slider.value for slider in self.sliders}


class ChannelFilterView(HasWidget):

    def __init__(self, model: ChannelFilterModel):
        self.model = model

        widget = QWidget()
        HasWidget.__init__(self, widget=widget)

        self.layout = QVBoxLayout()
        widget.setLayout(self.layout)

        self.model.observe(self.render)

    def render(self, change=None):

        if change is None or len(change.old) != len(change.new):
            layout: QLayout = self.layout

            # Delete any existing sliders
            if (num_sliders := layout.count()) > 0:
                for idx in reversed(range(num_sliders)):
                    item: QWidgetItem = layout.itemAt(idx)
                    layout.removeItem(item)
                    item.widget().deleteLater()

            layout.update()
            assert layout.count() == 0  # should be no sliders in the layout at this point.

            # Make new sliders
            for slider_model in self.model.sliders:
                print(f"making slider from {slider_model}")
                slider = LabelledSliderView(model=slider_model)
                layout.addWidget(slider.widget)



from os import path

from PIL import ImageGrab
from PySide2.QtWidgets import QWidget


def save_screenshot(widget: QWidget, filename: str, use_pil=True) -> None:
    if use_pil:
        ImageGrab.grab(bbox=widget.geometry().getCoords()).save(filename)
    else:
        widget.grab().save(filename, path.splitext(filename)[-1][1:])



from typing import Optional

from bg_atlasapi import BrainGlobeAtlas
from treelib import Tree, Node


def create_brain_region_tree(atlas: Optional[BrainGlobeAtlas]) -> Tree:
    if atlas is None:
        return Tree()

    new_tree = Tree()
    for id in (tree := atlas.hierarchy).expand_tree(mode=Tree.DEPTH):
        region_name = atlas._get_from_structure(id, "name")
        node = Node(identifier=id, data=region_name)
        new_tree.add_node(node, parent=tree.parent(id))
    return new_tree


import sys

from PySide2.QtWidgets import QMessageBox


def _show_exception_dialog(t, value, traceback):
    QMessageBox.critical(None, "An Exception was Raised", f'Value: {value}')


def show_dialog_box_on_uncaught_exception():
    """
        Set sys.excepthook to this function so that a dialog box appears when exceptions are raised.
        from https://www.youtube.com/watch?v=hhRKiMceaeY
    """
    sys.excepthook = _show_exception_dialog




from dataclasses import dataclass
from typing import List, Optional

import numpy as np
import vedo
from traitlets import HasTraits, Instance
from vedo import Plotter, pyplot
from vtkmodules.qt.QVTKRenderWindowInteractor import QVTKRenderWindowInteractor

from regexport.model import AppState
from regexport.views.utils import HasWidget


@dataclass
class HistogramData:
    bin_edges: np.ndarray
    bar_heights: np.ndarray
    colors: List[str]
    x_labels: List[str]
    title: str = ""

    def __post_init__(self):
        assert self.bar_heights.ndim == 1
        assert len(self.bar_heights) == len(self.bin_edges) - 1
        assert len(self.bar_heights) == len(self.colors)
        assert len(self.bar_heights) == len(self.x_labels)


class HistogramModel(HasTraits):
    histogram = Instance(HistogramData, allow_none=True)

    def register(self, model: AppState):
        self.model = model
        model.observe(self.update, ['selected_cells', 'column_to_plot'])

    def update(self, change):
        model = self.model
        if model.selected_cells is None:
            self.histogram = None
        elif (data_column := model.selected_cells[model.column_to_plot]).dtype.name == 'category':
            self.histogram = None
        else:
            heights, bin_edges = np.histogram(data_column.values, bins='auto', density=True)
            self.histogram = HistogramData(
                bin_edges=bin_edges,
                bar_heights=heights,
                colors=['olivedrab'] * len(heights),
                x_labels=bin_edges[:-1].astype(int).astype(str).tolist(),
            )


class HistogramView(HasWidget):

    def __init__(self, model: HistogramModel):
        widget = QVTKRenderWindowInteractor()
        HasWidget.__init__(self, widget=widget)
        self.plotter = Plotter(qtWidget=widget)
        self.model = model
        self.model.observe(self.render)

    @staticmethod
    def render_histogram_data(data: HistogramData) -> vedo.pyplot.Plot:
        return vedo.pyplot.plot([data.bar_heights, data.x_labels, data.colors, data.bin_edges], mode='bars')

    def render(self, change=None):
        self.plotter.clear()
        hist: Optional[HistogramData] = self.model.histogram
        if hist is not None:
            hist_actor = self.render_histogram_data(data=hist)
            self.plotter.show(hist_actor, mode=12)



from PySide2.QtCore import Qt
from PySide2.QtWidgets import QWidget, QHBoxLayout, QLabel, QSlider
from traitlets import HasTraits, Unicode, Int, observe

from regexport.views.utils import HasWidget


class LabelledSliderModel(HasTraits):
    label = Unicode()
    min = Int(default_value=0)
    max = Int()
    value = Int(default_value=1000)
    label2 = Unicode(allow_none=True)

    def __repr__(self):
        return f"{self.__class__.__name__}(label='{self.label}', min={self.min}, max={self.max}, value={self.value})"

    @observe('value')
    def _clamp_value_to_be_inside_bounded_range(self, change):
        value = change.new
        print(self)
        if value < self.min:
            self.value = 0
        elif value > self.max:
            self.value = self.max


class LabelledSliderView(HasWidget):

    def __init__(self, model: LabelledSliderModel):
        widget = QWidget()
        HasWidget.__init__(self, widget=widget)

        layout = QHBoxLayout()
        widget.setLayout(layout)

        self.label = QLabel()
        layout.addWidget(self.label)

        self.slider = QSlider()
        self.slider.setOrientation(Qt.Horizontal)

        layout.addWidget(self.slider)

        self.value_label = QLabel()
        layout.addWidget(self.value_label)

        self.model = model
        self.model.observe(self.render)
        self.slider.valueChanged.connect(self._update_model_value)
        self.render()

    def render(self, change=None):
        self.label.setText(self.model.label)
        self.slider.setMinimum(self.model.min)
        self.slider.setMaximum(self.model.max)
        self.value_label.setText(str(self.model.value))

    def _update_model_value(self, value: int):
        self.model.value = value



from abc import ABC

from PySide2.QtWidgets import QWidget


class HasWidget(ABC):

    def __init__(self, widget: QWidget):
        self.__widget = widget

    @property
    def widget(self) -> QWidget:
        return self.__widget





from pathlib import Path
from typing import List

import pandas as pd
from PySide2.QtWidgets import QFileDialog, QAction
from bg_atlasapi import BrainGlobeAtlas
from traitlets import HasTraits, Unicode, directional_link, Instance, Bool

from regexport.model import AppState
from regexport.utils.load_data import read_detection_files


class LoadCellsActionModel(HasTraits):
    text = Unicode("2. Load TSV Files")
    atlas = Instance(BrainGlobeAtlas, allow_none=True)
    cells = Instance(pd.DataFrame, allow_none=True)
    enabled = Bool(default_value=False)

    def register(self, model: AppState):
        directional_link((model, 'atlas'), (self, 'atlas'))
        directional_link((self, 'cells'), (model, 'cells'))
        directional_link((model, 'atlas'), (self, 'enabled'), lambda atlas: atlas is not None)

    def submit(self, filenames: List[Path]):
        if not filenames:
            return
        if self.atlas is None:
            raise ValueError("No atlas detected, cannot register brain regions")
        df = read_detection_files(filenames, atlas=self.atlas)
        self.cells = df


class LoadCellsAction(QAction):

    def __init__(self, model: LoadCellsActionModel, *args, **kwargs):
        self.model = model
        super().__init__(*args, **kwargs)
        self.setText(self.model.text)

        self.triggered.connect(self.run)
        self.model.observe(self.set_enabled, 'enabled')
        self.set_enabled(None)
        
    def set_enabled(self, changed):
        print("setting enabled")
        self.setEnabled(self.model.enabled)

    def run(self):
        filenames, filetype = QFileDialog.getOpenFileNames(
            caption="Load Cell Points from File",
            filter="TSV Files (*.tsv)"
        )

        filenames = [Path(f) for f in filenames]
        self.model.submit(filenames=filenames)




from pathlib import Path
from typing import Tuple, List

import numpy as np
import pandas as pd
from bg_atlasapi import BrainGlobeAtlas


def read_detection_file(filename: Path, atlas: BrainGlobeAtlas) -> pd.DataFrame:
    df = pd.read_csv(filename, sep='\t')
    print(df.columns)
    x_mm_col = 'Allen CCFv3 X mm'
    y_mm_col = 'Allen CCFv3 Y mm'
    z_mm_col = 'Allen CCFv3 Z mm'
    return (
        df
            .astype({
                x_mm_col: float,
                y_mm_col: float,
                z_mm_col: float,
        })
            .pipe(lambda df: df.assign(
                BGIdx=get_brain_region_label(df[x_mm_col].values, df[y_mm_col].values,
                                             df[z_mm_col].values, annotation=atlas.annotation,
                                             resolution=atlas.resolution)))
            .merge(atlas.lookup_df, right_on="id", left_on="BGIdx", how="left")
            .rename(columns={
                x_mm_col: 'X',
                y_mm_col: 'Y',
                z_mm_col: 'Z',
                'name': 'BrainRegion',
                'acronym': 'Acronym',
        })
            .drop(columns=['id'])
            .assign(Image=filename.name)
    )  # .sample(10000)


def get_brain_region_label(x, y, z, annotation: np.ndarray, resolution: Tuple[int, int, int]) -> np.ndarray:
    sx, sy, sz = annotation.shape
    rx, ry, rz = resolution
    return annotation[
        np.clip((x / (rx / 1000)).astype(int), 0, sx - 1),
        np.clip((y / (ry / 1000)).astype(int), 0, sy - 1),
        np.clip((z / (rz / 1000)).astype(int), 0, sz - 1),
    ]


def read_detection_files(filenames: List[Path], atlas):
    df = pd.concat((read_detection_file(filename=filename, atlas=atlas) for filename in filenames), ignore_index=True)
    df = df.astype({'Acronym': 'category', 'BrainRegion': 'category', 'Image': 'category'})
    return df



from pathlib import Path

import pandas as pd
from PySide2.QtWidgets import QFileDialog, QCheckBox, QDialog, QAction
from traitlets import HasTraits, Bool, directional_link

from regexport.model import AppState


class SaveCellsActionModel(HasTraits):
    text = "3. Save Cells"
    enabled = Bool(default_value=False)

    def register(self, model: AppState):
        self.model = model
        directional_link((model, 'cells'), (self, 'enabled'), lambda cells: cells is not None)

    def submit(self, filename: Path, selected_regions_only: bool = False):
        print('File saving...')

        df = self.model.selected_cells if selected_regions_only else self.model.cells
        types = {
            'Image': 'category',
            'BrainRegion': 'category',
            'Acronym': 'category',
            'X': 'float32',
            'Y': 'float32',
            'Z': 'float32',
        }

        types.update({col: 'uint16' for col in self.model.cells.columns if "Num Spots" in col})
        df = df.astype(types)
        df: pd.DataFrame = df.drop(columns=['BGIdx'])

        print(df.info())
        print(filename)

        if filename.suffix.lower() == ".csv":
            df.to_csv(filename, index=False)
        elif filename.suffix.lower() == ".feather":
            df.reset_index(drop=True).to_feather(filename)
        else:
            raise TypeError(f"Error saving file {str(filename)}: {filename.suffix} extension not supported.")
        print("File saved")


class ChkBxFileDialog(QFileDialog):
    def __init__(self, chkBxTitle="Selected Regions Only", filter="*.txt"):
        super().__init__(filter=filter)
        self.setSupportedSchemes(["file"])
        self.setOption(QFileDialog.DontUseNativeDialog)
        self.setAcceptMode(QFileDialog.AcceptSave)
        self.setNameFilter("Feather file (*.feather);;CSV file (*.csv)")
        self.selectNameFilter("Feather file (*.feather);;CSV file (*.csv)")
        self.chkBx = QCheckBox(chkBxTitle)
        self.layout().addWidget(self.chkBx)

    @property
    def full_filename(self) -> Path:
        filename = self.selectedUrls()[0].toLocalFile()
        extension_filter = self.selectedNameFilter()
        extension = extension_filter[extension_filter.index('*.') + 1:-1]
        full_filename = Path(filename).with_suffix(extension)
        return full_filename

    @property
    def selected_regions_only(self) -> bool:
        return self.chkBx.isChecked()


class SaveCellsAction(QAction):

    def __init__(self, model: SaveCellsActionModel, *args, **kwargs):
        self.model = model
        super().__init__(*args, **kwargs)

        self.setText(model.text)
        self.triggered.connect(self.click)
        self.model.observe(self.set_enabled, 'enabled')
        self.set_enabled(None)

    def set_enabled(self, changed):
        self.setEnabled(self.model.enabled)

    def click(self):
        dialog = ChkBxFileDialog()
        if dialog.exec_() == QDialog.Accepted:
            self.model.submit(
                filename=dialog.full_filename,
                selected_regions_only=dialog.selected_regions_only
            )



from typing import Tuple

from PySide2.QtWidgets import QVBoxLayout, QWidget

from .utils import HasWidget


class Sidebar(HasWidget):

    def __init__(self, widgets: Tuple[HasWidget, ...] = ()):

        self._widget = QWidget()
        HasWidget.__init__(self, widget=self._widget)

        layout = QVBoxLayout()
        self._widget.setLayout(layout)

        self.widgets = widgets
        for widget in widgets:
            layout.addWidget(widget.widget)








from importlib.resources import read_text
from pathlib import Path

from PySide2.QtWidgets import QFileDialog, QAction

from regexport import qupath_scripts


class SaveGroovyScriptActionModel:
    text = "0. Get QuPath Exporter"

    def submit(self, filename):
        groovy_script = read_text(qupath_scripts, 'export_registered_cells_to_tsv.groovy')
        with open(filename, 'w') as f:
            f.write(groovy_script)


class SaveGroovyScriptAction(QAction):

    def __init__(self, model: SaveGroovyScriptActionModel, *args, **kwargs):
        self.model = model
        super().__init__(*args, **kwargs)
        self.setText(model.text)
        self.triggered.connect(self.click)

    def click(self):
        filename, filetype_filter = QFileDialog.getSaveFileName(
            caption="Save QuPath Script",
            directory='export_registered_cells_to_tsv.groovy',
            filter="Groovy script (*.groovy)"
        )

        if not filename:
            return

        self.model.submit(Path(filename))



import sys
from pathlib import Path

from PySide2.QtWidgets import QApplication

from regexport.app import App
from regexport.utils.screenshots import save_screenshot


def main():
    print('starting app...')
    qapp = QApplication(sys.argv)
    app = App()
    win = app.create_gui()
    win.show()
    app.load_atlas_button.click()
    app.load_cells_button.submit([
        Path("example_data/tsvs_exported_from_qupath/section1.tsv"),
        Path("example_data/tsvs_exported_from_qupath/section2.tsv"),
    ])
    app.colordata_selector_dropdown.select("Esr1 (Opal 480): Num Spots")

    qapp.processEvents()
    save_screenshot(win, 'imgs/screenshot.jpg')

if __name__ == '__main__':
    main()



from PySide2.QtWidgets import QMainWindow

from regexport.actions.download_biop_extensions import SaveBiopExtensionsAction, SaveBiopExtensionsActionModel
from regexport.actions.load_atlas import LoadAtlasActionModel, LoadAtlasAction
from regexport.actions.load_cells import LoadCellsActionModel, LoadCellsAction
from regexport.actions.save_cells import SaveCellsActionModel, SaveCellsAction
from regexport.actions.save_script import SaveGroovyScriptActionModel, SaveGroovyScriptAction
from regexport.model import AppState
from regexport.utils.exceptions import show_dialog_box_on_uncaught_exception
from regexport.views.channel_filter import ChannelFilterView, ChannelFilterModel
from regexport.views.histogram import HistogramModel, HistogramView
from regexport.views.main_window import MainWindow
from regexport.views.plot_3d import PlotterModel, PlotterView
from regexport.views.region_tree import BrainRegionTreeModel, BrainRegionTree
from regexport.views.sidebar import Sidebar
from regexport.views.text_selector import TextSelectorModel, DropdownTextSelectorView


class App:
    def __init__(self):
        self.model = AppState()

        self.plot_window = PlotterModel()
        self.plot_window.register(model=self.model)

        self.brain_region_tree = BrainRegionTreeModel()
        self.brain_region_tree.register(model=self.model)

        self.load_atlas_button = LoadAtlasActionModel()
        self.load_atlas_button.register(model=self.model)

        self.load_cells_button = LoadCellsActionModel()
        self.load_cells_button.register(model=self.model)

        self.export_data_button = SaveCellsActionModel()
        self.export_data_button.register(model=self.model)

        self.colormap_selector_model = TextSelectorModel()
        self.colormap_selector_model.register(model=self.model, options_attr='colormap_options',
                                              selected_attr='selected_colormap')

        self.colordata_selector_dropdown = TextSelectorModel()
        self.colordata_selector_dropdown.register(model=self.model, options_attr='column_to_plot_options',
                                                  selected_attr='column_to_plot')

        self.num_spots_histogram = HistogramModel()
        self.num_spots_histogram.register(model=self.model)

        self.download_biop_extensions_button = SaveBiopExtensionsActionModel()
        self.save_groovy_script_button = SaveGroovyScriptActionModel()

        self.channel_filter_model = ChannelFilterModel()
        self.channel_filter_model.register(model=self.model)

    def create_gui(self) -> QMainWindow:
        show_dialog_box_on_uncaught_exception()
        return MainWindow(
            main_widgets=(
                BrainRegionTree(model=self.brain_region_tree),
                PlotterView(model=self.plot_window),
                Sidebar(
                    widgets=(
                        DropdownTextSelectorView(model=self.colordata_selector_dropdown),
                        DropdownTextSelectorView(model=self.colormap_selector_model),
                        HistogramView(model=self.num_spots_histogram),
                        ChannelFilterView(model=self.channel_filter_model),
                    )
                ),
            ),
            menu_actions=(
                SaveBiopExtensionsAction(model=self.download_biop_extensions_button),
                SaveGroovyScriptAction(model=self.save_groovy_script_button),
                LoadAtlasAction(model=self.load_atlas_button),
                LoadCellsAction(model=self.load_cells_button),
                SaveCellsAction(model=self.export_data_button),
            )
        )

    def set_max(self, channel_name, max_spots):
        ...



from pathlib import Path
from typing import Optional

import numpy as np
from traitlets import HasTraits, Instance, directional_link
from vedo import Plotter, Mesh, Points
from vtkmodules.qt.QVTKRenderWindowInteractor import QVTKRenderWindowInteractor

from regexport.model import AppState
from regexport.utils.point_cloud import PointCloud
from regexport.utils.profiling import warn_if_slow
from regexport.views.utils import HasWidget


class PlotterModel(HasTraits):
    atlas_mesh = Instance(Path, allow_none=True)
    points = Instance(PointCloud, default_value=PointCloud())

    def register(self, model: AppState):
        self.model = model
        directional_link(
            (model, 'atlas'),
            (self, 'atlas_mesh'),
            lambda atlas: Path(str(atlas.structures[997]['mesh_filename'])) if atlas is not None else None
        )
        model.observe(self.link_cells_to_points, names=[
            'selected_cells', 'selected_colormap', 'column_to_plot',
        ])

    def link_cells_to_points(self, change):
        model = self.model
        if model.selected_cells is None:
            self.points = PointCloud()
            return
        color_col = model.selected_cells[model.column_to_plot]
        points = PointCloud.from_cmap(
            positions=model.selected_cells[['X', 'Y', 'Z']].values * 1000,
            color_levels=color_col.cat.codes.values if color_col.dtype.name == 'category' else color_col.values,
            cmap=self.model.selected_colormap
        )
        self.points = points


class PlotterView(HasWidget):

    def __init__(self, model: PlotterModel):
        self.model = model
        self.item_points = {}
        self._atlas_mesh = None

        widget = QVTKRenderWindowInteractor()
        HasWidget.__init__(self, widget=widget)
        self.plotter = Plotter(qtWidget=widget)

        self.model.observe(self.observe_atlas_mesh, ['atlas_mesh'])
        self.model.observe(self.render, ['points'])

    @property
    def atlas_mesh(self) -> Optional[Mesh]:
        return self._atlas_mesh

    @atlas_mesh.setter
    def atlas_mesh(self, value: Optional[Mesh]):
        self._atlas_mesh = value
        self.render(None)

    @staticmethod
    def load_mesh(filename: Path) -> Mesh:
        return Mesh(
            str(filename),
            alpha=0.1,
            computeNormals=True,
            c=(1., 1., 1.)
        )

    def observe_atlas_mesh(self, change):
        print('saw atlas change')
        if self.model.atlas_mesh is None:
            self._atlas_mesh = Mesh()
        else:
            print('loading')
            # worker = Task(self.load_mesh, self.model.atlas_mesh)
            # worker.signals.finished.connect(partial(setattr, self, "atlas_mesh"))
            # 
            # pool = QThreadPool.globalInstance()
            # pool.start(worker)
            self.atlas_mesh = self.load_mesh(self.model.atlas_mesh)


    @warn_if_slow()
    def render(self, change=None):
        actors = [self._atlas_mesh]
        # box = self._atlas_mesh.box().wireframe().alpha(0.2).color((255, 0, 0))

        # actors.append(box)
        if len((points := self.model.points).coords) > 0:
            coords = points.coords
            colors = (np.hstack((points.colors, points.alphas)) * 255).astype(int)  # alphas needed for fast rendering.
            actors.append(Points(coords, r=3, c=colors))

        self.plotter.clear(at=0)
        self.plotter.show(actors, at=0)
        # self.plotter.addInset(self._atlas_mesh, pos=(.9, .9), size=0.1, c='w', draggable=True)
        # note: look at from vedo.applications import SlicerPlotter for inspiration



import sys

from PySide2.QtWidgets import QApplication
#
from regexport.app import App
#

def main():
    qapp = QApplication(sys.argv)
    app = App()
    win = app.create_gui()
    win.show()
    sys.exit(qapp.exec_())


if __name__ == '__main__':
    main()



from pathlib import Path
from typing import Union, Tuple

from PIL import ImageGrab
from PySide2.QtCore import Qt, QThreadPool
from PySide2.QtWidgets import QMainWindow, QWidget, QSplitter, QToolBar, QAction

from .utils import HasWidget


class MainWindow(QMainWindow):

    def __init__(self, main_widgets: Tuple[Union[QWidget, HasWidget], ...] = (),
                 menu_actions: Tuple[QAction, ...] = ()):
        super().__init__()
        self._widgets = main_widgets  # keep reference to these widgets, so they aren't garbage collected.

        self._splitter = QSplitter(Qt.Horizontal)
        self.setCentralWidget(self._splitter)
        for widget in self._widgets:
            self._splitter.addWidget(widget.widget if hasattr(widget, 'widget') else widget)

        self._actions = menu_actions
        tool_bar = QToolBar()
        for action in menu_actions:
            tool_bar.addAction(action)
        self.addToolBar(Qt.TopToolBarArea, tool_bar)




from typing import Tuple, Any

from treelib import Tree


def is_parent(id: Any, selected_ids: Tuple[Any, ...], tree: Tree) -> bool:
    if id in selected_ids:
        return True
    else:
        return any(tree.is_ancestor(selected_id, id) for selected_id in selected_ids)



from enum import Enum, auto

import numpy as np
import pandas as pd
from bg_atlasapi import BrainGlobeAtlas
from matplotlib import pyplot as plt
from traitlets import HasTraits, Instance, observe, Tuple, List, Unicode, Dict, Int

from regexport.utils.filters import is_parent


class AnalysisType(Enum):
    RegionLabel = auto()
    SubCellCount = auto()


class AppState(HasTraits):
    atlas = Instance(BrainGlobeAtlas, allow_none=True)
    cells = Instance(pd.DataFrame, allow_none=True)
    selected_region_ids = Tuple(default_value=())  # should be tuple of ints
    selected_cell_ids = Instance(np.ndarray, default_value=np.array([], dtype=int))  # array of ints
    column_to_plot_options = List(Unicode(), default_value=["BrainRegion"])
    column_to_plot = Unicode(default_value="BrainRegion")
    colormap_options = List(Unicode(), default_value=[cmap for cmap in plt.colormaps() if not cmap.endswith('_r')])#['tab20c', 'viridis'])
    selected_colormap = Unicode(default_value='tab20c')
    selected_cells = Instance(pd.DataFrame, allow_none=True)
    max_numspots_filters = Dict(key_trait=Unicode(), value_trait=Int(), default_value={})

    @observe('cells')
    def _update_max_numspots_filter_to_max_of_each_numspots_column(self, change):
        if self.cells is None:
            self.max_numspots_filters = {}
        else:
            self.max_numspots_filters = {col: int(self.cells[col].max()) for col in self.cells.columns if 'Num Spots' in col}


    @observe('cells')
    def _update_column_to_plot_options(self, change):
        if self.cells is None:
            return
        self.cells: pd.DataFrame
        columns = [name for name in self.cells.columns if 'Num Spots' in name]
        self.column_to_plot = "BrainRegion"
        self.column_to_plot_options = ["BrainRegion"] + columns

    @observe('column_to_plot')
    def _validate_column_to_plot(self, change):
        assert self.column_to_plot in self.column_to_plot_options

    @observe('selected_region_ids', 'cells', 'max_numspots_filters')
    def _update_selected_cell_ids(self, change):

        if self.cells is None:
            return
        if len(self.selected_region_ids) == 0:
            self.selected_cell_ids = self.cells.index.values
            selected_cells = self.cells
        else:
            is_parented = self.cells.groupby('BGIdx', as_index=False).BGIdx.transform(
                lambda ids: is_parent(ids.values[0], selected_ids=self.selected_region_ids, tree=self.atlas.hierarchy) if ids.values[0] != 0 else False
            )
            only_parented = is_parented[is_parented.BGIdx].index.values
            self.selected_cell_ids = only_parented
            selected_cells = self.cells.iloc[only_parented]


        query = ' & '.join(f"(`{channel}` <= {value})" for channel, value in self.max_numspots_filters.items())

        print('querying')
        selected_cells2 = selected_cells.query(query)
        print('queried')
        self.selected_cells = selected_cells2



from typing import Callable

from PySide2.QtCore import QObject, Signal, QRunnable, QThread, Slot


class TaskSignals(QObject):
    finished = Signal(object)


class Task(QRunnable):

    def __init__(self, fun: Callable, *args, **kwargs):
        """
        Runs function with args and kwargs when Task.run() is called, outputting to the Task.signals.finished signal.

        Meant to be used with QThreadPool.start(my_task)

        Resources:
          - https://www.pythonguis.com/tutorials/multithreading-pyqt-applications-qthreadpool/
          - https://realpython.com/python-pyqt-qthread/#using-qthread-to-prevent-freezing-guis
        """

        super().__init__()
        self._orig_thread = QThread.currentThread()
        self.fun = fun
        self.args = args
        self.kwargs = kwargs
        self.signals = TaskSignals()

    def run(self):
        assert self._orig_thread != QThread.currentThread(), "Worker not running in seperate thread, pointless."
        result = self.fun(*self.args, **self.kwargs)
        self.signals.finished.emit(result)



import time
from _warnings import warn


def warn_if_slow(limit: float = 2):
    def decorator(fun):
        def wrapper(*args, **kwargs):
            start_time = time.monotonic()
            result = fun(*args, **kwargs)
            run_duration = time.monotonic() - start_time
            if run_duration > limit:
                warn(f"Slow Run ({run_duration:.2} secs): {fun}")
            return  result
        return wrapper
    return decorator


def since():
    nt = time.monotonic()
    yield 0.
    while True:
        ot, nt = nt, time.monotonic()
        yield nt - ot


from functools import partial

from PySide2.QtCore import QThreadPool
from PySide2.QtWidgets import QAction
from bg_atlasapi import BrainGlobeAtlas
from traitlets import HasTraits, Unicode, Instance, directional_link

from regexport.model import AppState
from regexport.utils.parallel import Task


class LoadAtlasActionModel(HasTraits):
    text = Unicode("1. Load Brainglobe Atlas")
    atlas = Instance(BrainGlobeAtlas, allow_none=True)

    def register(self, model: AppState):
        directional_link((self, "atlas"), target=(model, "atlas"))

    def click(self):
        self.atlas = self.load_brainglobe_atlas()

    @staticmethod
    def load_brainglobe_atlas() -> BrainGlobeAtlas:
        # Static method used for multithreading
        return BrainGlobeAtlas("allen_mouse_25um")


class LoadAtlasAction(QAction):

    def __init__(self, model: LoadAtlasActionModel, *args, **kwargs):
        self.model = model
        super().__init__(*args, **kwargs)
        self.setText(self.model.text)
        self.triggered.connect(self.click)

    def click(self):
        task = Task(self.model.load_brainglobe_atlas)
        task.signals.finished.connect(partial(setattr, self.model, 'atlas'))
        QThreadPool.globalInstance().start(task)




from PySide2.QtWidgets import QTreeWidget, QAbstractItemView, QTreeWidgetItem
from traitlets import HasTraits, Instance, Tuple, directional_link
from treelib import Tree

from regexport.model import AppState
from regexport.utils.atlas import create_brain_region_tree
from .utils import HasWidget


class BrainRegionTreeModel(HasTraits):
    tree = Instance(Tree, allow_none=False, default_value=Tree())
    selected_region_ids = Tuple(default_value=())  # Tuple of ints

    def register(self, model: AppState):
        directional_link((self, 'selected_region_ids'), (model, 'selected_region_ids'))
        directional_link((model, 'atlas'), (self, 'tree'), create_brain_region_tree)

    def select(self, *brain_region_ids: int):
        print(brain_region_ids)
        self.selected_region_ids = brain_region_ids


class BrainRegionTree(HasWidget):

    def __init__(self, model: BrainRegionTreeModel):
        self.model = model

        treeview = QTreeWidget()
        treeview.setHeaderHidden(True)
        treeview.setWordWrap(False)
        treeview.setSelectionMode(QAbstractItemView.ExtendedSelection)
        treeview.setSelectionBehavior(QAbstractItemView.SelectRows)
        treeview.itemSelectionChanged.connect(self.onSelectionChanged)
        HasWidget.__init__(self, widget=treeview)

        self.treeview = treeview

        self.model.observe(self.render, names=['tree'])

    def render(self, change=None):
        tree = self.model.tree

        # no need to render empty tree
        if len(tree) == 0:
            return

        ids = tree.expand_tree(mode=Tree.DEPTH)
        next(ids)  # skip displaying root

        for id in ids:
            node = tree.get_node(id)
            node.item = QTreeWidgetItem()
            node.item.setText(0, node.data)
            node.item.setText(1, str(node.identifier))

            if (parent := tree.parent(node.identifier)).is_root():
                self.treeview.addTopLevelItem(node.item)
            else:
                parent.item.addChild(node.item)

        # Finish up
        self.treeview.expandToDepth(1)

    def onSelectionChanged(self):
        self.model.select(*[int(item.text(1)) for item in self.treeview.selectedItems()])




from __future__ import annotations

from dataclasses import dataclass, field

import numpy as np
from matplotlib import pyplot as plt
from matplotlib.colors import ListedColormap


@dataclass(frozen=True)
class PointCloud:
    coords: np.ndarray = field(default=np.empty((0, 3), dtype=float))
    colors: np.ndarray = field(default=np.empty((0, 3), dtype=float))
    alphas: np.ndarray = field(default=np.empty((0, 1), dtype=float))

    def __post_init__(self):
        assert self.coords.ndim == 2 and self.coords.shape[1] == 3
        assert self.colors.ndim == 2 and self.colors.shape[1] == 3
        assert self.alphas.ndim == 2 and self.alphas.shape[1] == 1


    @classmethod
    def from_cmap(cls, positions: np.ndarray, color_levels: np.ndarray, cmap: str = 'tab20c') -> PointCloud:
        selected_colors = convert_values_to_colors(color_levels, getattr(plt.cm, cmap))
        return cls(
            coords=positions,
            colors=selected_colors[:, :3],
            alphas=selected_colors[:, 3:4],
        )


def convert_values_to_colors(color_codes: np.ndarray, cmap: ListedColormap):
    return cmap(color_codes / color_codes.max())[:, :4]
